package mue.test;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import robocode.control.RobotResults;

/**
 * Accumulates the results of battles with same participants. 
 * 
 * @author mue
 */
public class BattleResults {
	
	/**
	 * Used to cache and sum up battle results for one robot.
	 */
	protected static class BattleData {
		private int score;
		private double scorePercentage;
		private int survival;
		private int lastSurvivorBonus;
		private int bulletDamage;
		private int bulletDamageBonus;
		private int ramDamage;
		private int ramDamageBonus;
		private int firsts;
		private int seconds;
		private int thirds;
		private List scorePercentages;
		
		/**
		 * Creates empty battle data.
		 */
		public BattleData() {
			score = 0;
			scorePercentage = 0;
			survival = 0;
			lastSurvivorBonus = 0;
			bulletDamage = 0;
			bulletDamageBonus = 0;
			ramDamage = 0;
			ramDamageBonus = 0;
			firsts = 0;
			seconds = 0;
			thirds = 0;
			scorePercentages = new ArrayList();
		}
		
		/**
		 * Adds the specified result to this battle data.
		 * 
		 * @param result the result 
		 * @param newScorePercentage the percentage of the total score the corresponding robot archieved
		 */
		public void addResult(RobotResults result, int maxScore) {
			score += result.getScore();
			Double percentage = new Double((double)(result.getScore() * 100) / maxScore);
			scorePercentages.add(percentage);
			scorePercentage += percentage.doubleValue();
			survival += result.getSurvival();
			lastSurvivorBonus += result.getLastSurvivorBonus();
			bulletDamage += result.getBulletDamage();
			bulletDamageBonus += result.getBulletDamageBonus();
			ramDamage += result.getRamDamage();
			ramDamageBonus += result.getRamDamageBonus();
			firsts += result.getFirsts();
			seconds += result.getSeconds();
			thirds += result.getThirds();
		}
		
		/**
		 * Returns a String containing the battle data.
		 * 
		 * @param maxScore the maximum score that could be archieved in all battles
		 * @param battles the number of battles this data was recorded in 
		 * @return a String containing the battle data
		 */
		public String getResults(int newMaxScore,int battles) {
			StringBuffer buffer = new StringBuffer();
			double averageScorePercentage = scorePercentage / battles;
			Iterator iterator = scorePercentages.iterator();
			double minPercentage = 100;
			double maxPercentage = 0;
			double avgDeviation = 0;
			double stdDeviation = 0;
			while (iterator.hasNext()) {
				Double percentage = (Double)iterator.next();
				if (percentage.doubleValue() < minPercentage) minPercentage = percentage.doubleValue();
				if (percentage.doubleValue() > maxPercentage) maxPercentage = percentage.doubleValue();
				avgDeviation += Math.abs(percentage.doubleValue() - averageScorePercentage);
				stdDeviation += (percentage.doubleValue() - averageScorePercentage) * (percentage.doubleValue() - averageScorePercentage);
			}
			avgDeviation = (double)Math.round(avgDeviation * 100 / battles) / 100;
			minPercentage = (double)Math.round(minPercentage * 100) / 100;
			maxPercentage = (double)Math.round(maxPercentage * 100) / 100;
			if (battles < 2) {
				stdDeviation = 0;
			} else {
				stdDeviation = (double)Math.round(100 * Math.sqrt(stdDeviation / ((battles - 1) * (battles - 1)))) / 100;
			} 
			averageScorePercentage = (double)Math.round(100 * averageScorePercentage) / 100;
			buffer.append("  percentage:      ").append(averageScorePercentage).append("%");
			buffer.append(" (std. deviation: ").append(stdDeviation).append("%)\n");
			buffer.append("                          (min: ").append(minPercentage).append("%, max: ").append(maxPercentage);
			buffer.append("%, avg. deviation: ").append(avgDeviation).append("%)\n"); 
			buffer.append("  score:           ").append(score);
			buffer.append(" (relative: ").append(percentageString(score,newMaxScore)).append(") \n");
			buffer.append("  survival:        ").append(survival);
			buffer.append(" (bonus: ").append(lastSurvivorBonus).append(")\n");
			buffer.append("  bullet damage:   ").append(bulletDamage);			
			buffer.append(" (bonus: ").append(bulletDamageBonus).append(")\n");
			buffer.append("  ram damage (*2): ").append(ramDamage * 2);			
			buffer.append(" (bonus: ").append(ramDamageBonus).append(")\n");
			buffer.append("  survivals:       ").append("first: ").append(firsts);
			buffer.append(", second: ").append(seconds);
			buffer.append(", third: ").append(thirds).append("\n");
			return buffer.toString();
		}
		
		/**
		 * Returns the sum of the score percentages this robot archieved.
		 * 
		 * @return the sum of the score percentages
		 */
		public double getScorePercentage() {
			return scorePercentage;
		}
		
		/**
		 * Returns the variance of the score percentages this robot archieved.
		 * 
		 * @param battles the number of battles this data is about
		 * @return the variance of the score percentages this robot archieved
		 */
		public double getVariance(int battles) {
			double averageScorePercentage = scorePercentage / battles;
			Iterator iterator = scorePercentages.iterator();
			double variance = 0;
			while (iterator.hasNext()) {
				Double percentage = (Double)iterator.next();
				variance += (percentage.doubleValue() - averageScorePercentage) * (percentage.doubleValue() - averageScorePercentage);
			}
			return (battles < 2) ? 0 : variance / (battles - 1);
		}
		
		/**
		 * Returns a string containing the percentage of points from maxPoints.
		 * 
		 * @param points the points the percentage is calculated with
		 * @param maxPoints the maximum number of points the percentage is calculated with
		 * @return a string containing the percentage of points from maxPoints
		 */
		protected String percentageString(long points, long maxPoints) {
			if (maxPoints == 0) return " 0.00%";
			long percentage = (points*10000) / maxPoints;
			return "" + (percentage/100) + "." + ((percentage/10)%10) + "" + (percentage%10) + "%";
		}
	}
	
	protected Map results;
	protected int rounds;
	protected int maxScore;
	protected int battles;
	

	/**
	 * Creates empty battle results.
	 */
	public BattleResults() {
		results = new TreeMap();
		rounds = 0;
		battles = 0;
		maxScore = 0;
	}
	
	/**
	 * Adds the specified results to this data.
	 *  
	 * @param newRounds the number of rounds of the battle.
	 * @param robotResults the results to be added
	 */
	public void addResults(int newRounds, RobotResults robotResults[]) {
		rounds += newRounds;
		battles++;
		int totalBattleScore = 0;
		for (int i = 0; i < robotResults.length; i++) totalBattleScore += robotResults[i].getScore();
		for (int i = 0; i < robotResults.length; i++) {
			BattleData battleData = (BattleData)results.get(RobocodeListener.getRobotId(robotResults[i].getRobot()));
			if (battleData == null) {
				battleData = new BattleData();
				results.put(RobocodeListener.getRobotId(robotResults[i].getRobot()),battleData);
			}
			battleData.addResult(robotResults[i],totalBattleScore);
		}
		maxScore += totalBattleScore;
	}
	
	/**
	 * Returns a string containing these battle results.
	 * 
	 * @return a string containing these battle results
	 */
	public String getResults() {
		Map order = new TreeMap();
		StringBuffer buffer = new StringBuffer("result of ");
		Iterator iterator = results.keySet().iterator();
		while (iterator.hasNext()) {
			String key = (String)iterator.next();
			buffer.append(key);
			order.put(new Double(100 * battles - ((BattleData)results.get(key)).getScorePercentage()),key);
			if (iterator.hasNext()) buffer.append(" vs ");
		}
		buffer.append(" (").append(rounds).append(" rounds in ").append(battles).append(" battles):\n");
		
		iterator = order.keySet().iterator();
		while (iterator.hasNext()) {
			String key = (String)order.get(iterator.next());
			buffer.append(key).append(":\n");
			BattleData battleData = (BattleData)results.get(key);
			buffer.append(battleData.getResults(maxScore,battles));
		}		
		buffer.append("\n");
		return buffer.toString();
	}
	
	/**
	 * Returns the average score percentage specified robot got.
	 * 
	 * @param robot the name of the robot 
	 * @return the average score percentage specified robot got
	 */
	public double getAverageScorePercentage(String robot) {
		BattleData battleData = (BattleData)results.get(robot);
		if (battleData == null)
			throw new IllegalArgumentException("Nothing known about specified robot (" + robot + ")");
		if (battles == 0) return 0;
		return battleData.getScorePercentage() / battles;
	}
	
	/**
	 * Returns the variance of the score percentages the specified robot got
	 * 
	 * @param robot the name of the robot
	 * @return the variance of the score percentages the specified robot got
	 */
	public double getVariance(String robot) {
		BattleData battleData = (BattleData)results.get(robot);
		if (battleData == null)
			throw new IllegalArgumentException("Nothing known about specified robot (" + robot + ")");
		return battleData.getVariance(battles);
	}
	
	/**
	 * Returns the number of battles this result corresponds to.
	 * 
	 * @return the number of battles this result corresponds to
	 */
	public int getNumberOfBattles() {
		return battles;
	}
}