package mue.test;

import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import robocode.control.BattleSpecification;
import robocode.control.RobotResults;
import robocode.control.RobotSpecification;

/**
 * The listener that is passed to the robocode engine.
 * 
 * @author mue
 */
public class RobocodeListener implements robocode.control.RobocodeListener {
	
	protected boolean printEvents;
	protected boolean printMessages;
	protected Map battles;
	
	/**
	 * Returns an id for given robot specification.
	 * 
	 * @param robot the robot an id is created for
	 * @return an id for given robot specification
	 */
	public static String getRobotId(RobotSpecification robot) {
		String version = robot.getVersion();
		if ((version == null) || (version.trim().length() == 0)) return robot.getClassName();
		if (version.endsWith("*")) version = version.substring(0,version.length() - 1);
		return robot.getClassName() + " " + version;
	}

	/**
	 * Creates a listener with given parameters. The results are written to System.out
	 *
	 * @param newPrintEvents if true the events recieved are written to System.out
	 * @param newPrintMessages if true the messages received are written to System.out 
	 */
	public RobocodeListener(boolean newPrintEvents,boolean newPrintMessages) {
		printEvents = newPrintEvents;
		printMessages = newPrintMessages;
		battles = new HashMap();
	}

	/**
	 * This method is called if the battle was aborted.
	 * 
	 * @param battle the specification of the battle aborted
	 */
	public synchronized void battleAborted(BattleSpecification battle) {
		try {
			if (printEvents) System.out.println("battle aborted");
		} finally {
			notifyAll();
		}
	}

	/**
	 * This method is called if the battle is completed.
	 * 
	 * @param battle the specification of the battle aborted
	 * @param results the results of the battle
	 */
	public synchronized void battleComplete(BattleSpecification battle, RobotResults[] results) {
		try {
			int maxScore = 0;
			Set key = new HashSet(results.length * 2); 
			for (int i = 0; i < results.length; i++) {
				maxScore += results[i].getScore();
				key.add(getRobotId(results[i].getRobot()));
			} 
			BattleResults battleResults = (BattleResults)battles.get(key);
			if (battleResults == null) {
				battleResults = new BattleResults();
				battles.put(key,battleResults); 
			}
			battleResults.addResults(battle.getNumRounds(),results);
			
			if (printEvents) {
				System.out.print("battle completed: "); 
				for (int i = 0; i < results.length; i++) {
					String botName = results[i].getRobot().getClassName();
					String version = results[i].getRobot().getVersion();
					if (botName.lastIndexOf('.') < botName.length() - 1);
					botName = botName.substring(botName.lastIndexOf('.') + 1);
					if ((version != null) && (version.endsWith("*"))) version = version.substring(0,version.length() - 1);
					double score = (double)(Math.round(results[i].getScore() * 10000) / maxScore) / 100;
					System.out.print(botName + (version == null ? "" : " " + version) + " " + score + "%");
					if (i < results.length - 1) System.out.print("; ");
				}
				System.out.println();
			}
		} finally {
			notifyAll();			
		}
	}

	/**
	 * This method is called if a message was received.
	 * 
	 * @param message the message that is received
	 */
	public void battleMessage(String message) {
		if (printMessages) System.out.println("battle-message: " + message);
	}
	
	/**
	 * This method is called by RunBattles and waits until current battle is completed or aborted.
	 */
	public synchronized void waitForBattle() {
		try { 
			wait();
		} catch (InterruptedException e) {
			System.out.println("interrupted: " + e.getMessage());
		}
	}
	
	/**
	 * Writes the results of the battle with specified opponents to file or System out (if filename 
	 * is null).
	 * 
	 * @param robots the participants of the battle
	 * @param filename the filename or null
	 * @param append if true the result will be appended to the file, otherwise file is resetted
	 */
	public void write(Set robots, String filename, boolean append) {
		StringBuffer buffer = new StringBuffer();
		BattleResults battleResults = (BattleResults)battles.get(robots);
		buffer.append(battleResults.getResults());		
		write(buffer.toString(),filename,append);
	}
	
	/**
	 * Writes the evaluation of specified robot to file or System out (if filename is null).
	 * 
	 * @param robot the robot whose evaluation will be written
	 * @param filename the file the evaluation will be written to  
	 * @param append if true the result will be appended to the file, otherwise file is resetted
	 */
	public void writeEvaluation(String robot,String filename, boolean append) {
		StringBuffer buffer = new StringBuffer();
		
		double avgPercentage = 0;
		int numberOfBattles = 0;
		double stdDeviation = 0;
		
		Iterator iterator = battles.keySet().iterator();
		while (iterator.hasNext()) {
			Set robots = (Set)iterator.next();
			if (robots.contains(robot)) {
				BattleResults results = (BattleResults)battles.get(robots);
				avgPercentage += results.getAverageScorePercentage(robot) * results.getNumberOfBattles();
				stdDeviation += results.getVariance(robot) * (results.getNumberOfBattles() - 1);
				numberOfBattles += results.getNumberOfBattles();
			}
		}
		
		avgPercentage = numberOfBattles == 0 ? 0 : (double)Math.round(avgPercentage * 10000 / numberOfBattles) / 10000;
		if (numberOfBattles < 2) {
			stdDeviation = 0;
		} else {
			stdDeviation = (double)Math.round(10000 * Math.sqrt(stdDeviation / ((numberOfBattles - 1) * (numberOfBattles - 1)))) / 10000;
		}
		
		buffer.append("evaluation: ").append(avgPercentage).append("%; ");
		buffer.append("std. deviation: ").append(stdDeviation).append("%");
		buffer.append(" (battles: ").append(numberOfBattles).append(")\n");
		write(buffer.toString(),filename,append);
	}
	
	/**
	 * Writes the specified string to file or to System out (if filename is null).
	 * 
	 * @param text the string to be written
	 * @param filename the filename the text will be written to or null
	 * @param append if true the text will be appended to the file, otherwise file is resetted
	 */
	protected void write(String text, String filename, boolean append) {
		if (filename != null) {
			FileWriter output = null;
			try {
				output = new FileWriter(filename,append);
				output.write(text);
			} catch (IOException e) {
				System.out.println("error during output to file");
			} finally {
				if (output != null) try {output.close();} catch (IOException e) {}
			}
		} else {
			System.out.println(text); 
		}
	}
}
